from __future__ import print_function
import argparse
import os
import numpy
import random
import torch
import torch.nn as nn
import torch.backends.cudnn as cudnn
import torch.optim as optim
import torchvision.transforms as transforms
import torchvision.utils as vutils
from torch.autograd import Variable
from itertools import chain
from utils.dataset import DATASET
from utils.ImagePool import ImagePool
from model.fixup_net import FIXUPNET
import torch.nn.functional as F
from utils.fitsFun import DATASET_fits
from utils.fitsFun import LoadSaveFits

parser = argparse.ArgumentParser(description='train pix2pix model')
parser.add_argument('--batchSize', type=int, default=4, help='with batchSize=1 equivalent to instance normalization.')
parser.add_argument('--ngf', type=int, default=64*2)
parser.add_argument('--ndf', type=int, default=64*2)
parser.add_argument('--niter', type=int, default=100000, help='number of iterations to train for')
parser.add_argument('--lr', type=float, default=0.0001, help='learning rate, default=0.0002')
parser.add_argument('--beta1', type=float, default=0.5, help='beta1 for adam. default=0.5')
parser.add_argument('--weight_decay', type=float, default=1e-4, help='weight decay in network D, default=1e-4')
parser.add_argument('--cuda', action='store_true', help='enables cuda')
parser.add_argument('--outf', default='checkpoints/', help='folder to output images and model checkpoints')
parser.add_argument('--manualSeed', type=int, help='manual seed')
parser.add_argument('--dataPath',default='./DATA/',help='image data')
parser.add_argument('--loadSize', type=int, default=19, help='scale image to this size')
parser.add_argument('--fineSize', type=int, default=32, help='random crop image to this size')
parser.add_argument('--flip', type=int, default=1, help='1 for flipping image randomly, 0 for not')
parser.add_argument('--input_nc', type=int, default=3721, help='channel number of input image')
parser.add_argument('--output_nc', type=int, default=3721, help='channel number of output image')
parser.add_argument('--kernelSize', type=int, default=3, help='random crop kernel to this size')
parser.add_argument('--G_AB', default='', help='path to pre-trained G_AB')
parser.add_argument('--save_step', type=int, default=1000, help='save interval')
parser.add_argument('--log_step', type=int, default=1000, help='log interval')
parser.add_argument('--loss_type', default='mse', help='GAN loss type, bce|mse default is negative likelihood loss')
parser.add_argument('--poolSize', type=int, default=50, help='size of buffer in lsGAN, poolSize=0 indicates not using history')
parser.add_argument('--lambda_idt', type=float, default=10.0, help='weight of cycle loss idt')
parser.add_argument("--gpus", default="0", type=str, help="gpu ids (default: 0)")
opt = parser.parse_args()
print(opt)

try:
    os.makedirs(opt.outf)
except OSError:
    pass

if opt.manualSeed is None:
    opt.manualSeed = random.randint(1, 10000)
print("Random Seed: ", opt.manualSeed)
random.seed(opt.manualSeed)
torch.manual_seed(opt.manualSeed)
if opt.cuda:
    print("=> use  gpu id: '{}'".format(opt.gpus))
    os.environ["CUDA_VISIBLE_DEVICES"] = opt.gpus
    if not torch.cuda.is_available():
            raise Exception("No GPU found or Wrong gpu id, please run without --cuda")

opt.seed = random.randint(1, 10000)
print("Random Seed: ", opt.seed)
torch.manual_seed(opt.seed)
cudnn.benchmark = True
##########      dataset fits  #############
datasetA = DATASET_fits(opt.dataPath,opt.fineSize)
loader_A= torch.utils.data.DataLoader(dataset=datasetA,
                                       batch_size=opt.batchSize,
                                       shuffle=True,
                                       num_workers=2)
loaderA = iter(loader_A)
# ABPool = ImagePool(opt.poolSize)
# BAPool = ImagePool(opt.poolSize)

############   MODEL   ###########
# custom weights initialization called on netG and netD
def weights_init(m):
    classname = m.__class__.__name__
    if classname.find('Conv') != -1:
        m.weight.data.normal_(0.0, 0.02)
    elif classname.find('BatchNorm') != -1:
        m.weight.data.normal_(1.0, 0.02)
        m.bias.data.fill_(0)

ndf = opt.ndf
ngf = opt.ngf

G_AB = FIXUPNET(opt.input_nc, opt.output_nc, opt.ngf)

if(opt.G_AB != ''):
    print('Warning! Loading pre-trained weights.')
    G_AB.load_state_dict(torch.load(opt.G_AB))
else:
    G_AB.apply(weights_init)
    

if(opt.cuda):
    G_AB.cuda()

###########   LOSS & OPTIMIZER   ##########
# criterionMSE = nn.MSELoss() 
criterionMSE = nn.L1Loss()
if(opt.loss_type == 'bce'):
    criterion = nn.BCELoss()
else:
    criterion = nn.MSELoss()
# chain is used to update two generators simultaneously
optimizerG = torch.optim.Adam(G_AB.parameters(),lr=opt.lr, betas=(opt.beta1, 0.999))

############   GLOBAL VARIABLES   ###########
input_nc = opt.input_nc
output_nc = opt.output_nc
fineSize = opt.fineSize
batchSize = opt.batchSize
kernelSize = opt.kernelSize
real_A = torch.FloatTensor(opt.batchSize, input_nc, fineSize, fineSize)
AB = torch.FloatTensor(opt.batchSize, input_nc, fineSize, fineSize)
real_B = torch.FloatTensor(opt.batchSize, output_nc, fineSize, fineSize)

real_A = Variable(real_A)
real_B = Variable(real_B)
AB = Variable(AB)

if(opt.cuda):

    real_A = real_A.cuda()
    real_B = real_B.cuda()
    AB = AB.cuda()
    criterion.cuda()
    criterionMSE.cuda()


lossdata = []
###########   Testing    ###########
def test(niter):
    loaderA = iter(loader_A)
    imgA = loaderA.next()
    real_A.resize_(imgA[:,0:input_nc,:,:].size()).copy_(imgA[:,0:input_nc,:,:])
    real_B.resize_(imgA[:,input_nc:input_nc+input_nc,:,:].size()).copy_(imgA[:,input_nc:input_nc+input_nc,:,:])
    AB = G_AB(real_A)
    
    LoadSaveFits.save_fit(AB.data[:,:,:,:],'AB_%03d_'%niter,'./out_picture/out_image_train/')
    LoadSaveFits.save_fit(real_A.data[:,:,:,:],'realA_%03d_'%niter,'./out_picture//out_image_train/')
    LoadSaveFits.save_fit(real_B.data[:,:,:,:],'realB_%03d_'%niter,'./out_picture//out_image_train/')  
###########   Training    ###########
G_AB.train()
for iteration in range(1,opt.niter+1):
    ###########   data  ###########
    try:
        imgA = loaderA.next()
    except StopIteration:
        loaderA = iter(loader_A)
        imgA = loaderA.next()

    real_A.resize_(imgA[:,0:input_nc,:,:].size()).copy_(imgA[:,0:input_nc,:,:])
    real_B.resize_(imgA[:,input_nc:input_nc+input_nc,:,:].size()).copy_(imgA[:,input_nc:input_nc+input_nc,:,:])

    ########### fGx ###########
    G_AB.zero_grad()
    AB = G_AB(real_A)  
    l_idt = criterionMSE(AB,real_B)*opt.lambda_idt

    errG = l_idt
    datanumber = torch.Tensor.cpu(errG.data)
    datanumber = datanumber.data.numpy()
    lossdata.append((datanumber.data).tolist())
    
    errG.backward()
    optimizerG.step()

    ###########   Logging   ############
    if(iteration % opt.log_step):
        print('[%d/%d] Loss_errG: %.4f '
                  % (iteration, opt.niter, errG.data ))
    ########## Visualize #########
    if(iteration % opt.log_step == 0):
        test(iteration)

    if iteration % opt.save_step == 0:
        torch.save(G_AB.state_dict(), '{}/G_AB_{}.pth'.format(opt.outf, iteration))
        filepath = './losstest02.txt'
        with open(filepath, 'w') as f:
            f.write(str(lossdata))



