import torch.nn as nn
import torch

# courtesy: https://github.com/darkstar112358/fast-neural-style/blob/master/neural_style/transformer_net.py
class InstanceNormalization(torch.nn.Module):
    """InstanceNormalization
    Improves convergence of neural-style.
    ref: https://arxiv.org/pdf/1607.08022.pdf
    """

    def __init__(self, dim, eps=1e-9):
        super(InstanceNormalization, self).__init__()
        self.scale = nn.Parameter(torch.FloatTensor(dim))
        self.shift = nn.Parameter(torch.FloatTensor(dim))
        self.eps = eps
        self._reset_parameters()

    def _reset_parameters(self):
        self.scale.data.uniform_()
        self.shift.data.zero_()

    def forward(self, x):
        n = x.size(2) * x.size(3)
        t = x.view(x.size(0), x.size(1), n)
        mean = torch.mean(t, 2,keepdim=True).unsqueeze(2).expand_as(x)
        # Calculate the biased var. torch.var returns unbiased var
        var = torch.var(t, 2,keepdim=True).unsqueeze(2).expand_as(x) * ((n - 1) / float(n))
        scale_broadcast = self.scale.unsqueeze(1).unsqueeze(1).unsqueeze(0)
        scale_broadcast = scale_broadcast.expand_as(x)
        shift_broadcast = self.shift.unsqueeze(1).unsqueeze(1).unsqueeze(0)
        shift_broadcast = shift_broadcast.expand_as(x)
        out = (x - mean) / torch.sqrt(var + self.eps)
        out = out * scale_broadcast + shift_broadcast
        return out
def conv3x3(in_planes, out_planes, stride=1):
    """3x3 convolution with padding"""
    return nn.Conv2d(in_planes, out_planes, kernel_size=3, stride=stride,
                     padding=1, bias=False)
def build_conv_block(dim):
    return nn.Sequential(nn.ReflectionPad2d((1,1,1,1)),
                         nn.Conv2d(dim,dim,3,1,0),
                         InstanceNormalization(dim),
                         # nn.BatchNorm2d(dim),
                         nn.ReLU(True),
                         nn.ReflectionPad2d((1,1,1,1)),
                         nn.Conv2d(dim,dim,3,1,0),
                         InstanceNormalization(dim)
                         # nn.BatchNorm2d(dim)
                         )
class FixupBasicBlock(nn.Module):
    expansion = 1

    def __init__(self, inplanes, stride=1, downsample=None):
        super(FixupBasicBlock, self).__init__()
        # Both self.conv1 and self.downsample layers downsample the input when stride != 1
        self.bias1a = nn.Parameter(torch.zeros(1))
        self.conv1 = conv3x3(inplanes, inplanes, stride)
        self.bias1b = nn.Parameter(torch.zeros(1))
        self.relu = nn.ReLU(inplace=True)
        self.bias2a = nn.Parameter(torch.zeros(1))
        self.conv2 = conv3x3(inplanes, inplanes)
        self.scale = nn.Parameter(torch.ones(1))
        self.bias2b = nn.Parameter(torch.zeros(1))
        self.downsample = downsample

    def forward(self, x):
        identity = x

        out = self.conv1(x + self.bias1a)
        out = self.relu(out + self.bias1b)

        out = self.conv2(out + self.bias2a)
        out = out * self.scale + self.bias2b

        if self.downsample is not None:
            identity = self.downsample(x + self.bias1a)
            identity = torch.cat((identity, torch.zeros_like(identity)), 1)

        out += identity
        out = self.relu(out)

        return out                        
class ResidualBlock(nn.Module):
    '''
    A single Residual Block:
    in ->
        conv(in_channels,out_channels,stride) -> BN -> ReLU
        conv(out_channels,out_channels,1) -> BN
    -> out
    (downsample)in + out
    '''
    def __init__(self, dim):
        super(ResidualBlock,self).__init__()
        self.conv = build_conv_block(dim)

    def forward(self, x):
        out = self.conv(x)
        out = out + x

        return out

class FIXUPNET(nn.Module):
    def __init__(self, input_nc, output_nc, ngf):
        super(FIXUPNET,self).__init__()
        # 128 x 128
        self.bias1 = nn.Parameter(torch.zeros(1))
        self.relu = nn.ReLU(inplace=True)

        self.layer1 = nn.Sequential( nn.ReflectionPad2d((1,1,1,1)),
                                     nn.Conv2d(input_nc,ngf,kernel_size=3,stride=1))
        self.layer2 = nn.Sequential(nn.Conv2d(ngf,ngf*2,kernel_size=3,stride=2,padding=1))
        # 64 x 64
        self.layer3 = nn.Sequential(nn.Conv2d(ngf*2,ngf*4,kernel_size=3,stride=2,padding=1))
        # 32 x 32
        self.layer4 = nn.Sequential(FixupBasicBlock(ngf*4),
                                    FixupBasicBlock(ngf*4),
                                    FixupBasicBlock(ngf*4),
                                    FixupBasicBlock(ngf*4),
                                    FixupBasicBlock(ngf*4),
                                    FixupBasicBlock(ngf*4),
                                    FixupBasicBlock(ngf*4),
                                    FixupBasicBlock(ngf*4))
        # 32 x 32
        self.layer5 = nn.Sequential(nn.ConvTranspose2d(ngf*4, ngf*2, 3, 2, 1, 1))
        # 64 x 64
        self.layer6 = nn.Sequential(nn.ConvTranspose2d(ngf*2, ngf*1, 3, 2, 1, 1))
        # 128 x 128
        self.layer7 = nn.Sequential(nn.ReflectionPad2d((1,1,1,1)),
                                     nn.Conv2d(ngf*1,output_nc,kernel_size=3,stride=1),
#                                     nn.Tanh()
                                     )
        # 128 x 128
    def forward(self,x):
        out = self.layer1(x)
        out = self.relu(out + self.bias1)
        out = self.layer2(out)
        out = self.relu(out + self.bias1)        
        out = self.layer3(out)
        out = self.relu(out + self.bias1)
        out = self.layer4(out)
        out = self.layer5(out)
        out = self.relu(out + self.bias1)
        out = self.layer6(out)
        out = self.relu(out + self.bias1)
        out = self.layer7(out)
        return out
